/**
 * @file
 * @copyright 2020, Valentin Anger
 * @license ISC
 */

import * as irc from "../net/irc/mod.ts";

export interface CommandModule {
  create(): Promise<Command>;
}

export interface Command {
  /** Command is admin only */
  readonly admin?: boolean;
  /** Command prefix */
  readonly command?: string;

  /** Command's description
   * Also used to determine whether it should be listed in help output */
  readonly description?: string;

  /** Called when Command is loaded
   * @param commander The Commander instance this Command is managed by
  */
  open?(commander: Commander): Promise<void>;
  /** Called before Command is unloaded
   * The Commander received during open should not be used after this function returns
  */
  close?(): Promise<void>;

  /** Called when client receives message with command prefix */
  handle?(message: irc.Message, argument: string | undefined): Promise<void>;

  /** Called when client joins channel */
  channelJoin?(message: irc.Message, channel: string): Promise<void>;
  /** Called when client leaves channel */
  channelLeave?(message: irc.Message, channel: string): Promise<void>;
  /** Called when nick joins channel */
  nickJoin?(
    message: irc.Message,
    channel: string,
    nickname: string,
  ): Promise<void>;
  /** Called when nick leaves channel */
  nickLeave?(
    message: irc.Message,
    channel: string,
    nickname: string,
  ): Promise<void>;
}

export class Commander {
  #commands: Map<string, Command> = new Map();
  /** Map channel + nickname to the last used command */
  #last_command: Map<string, Map<string, string>> = new Map();

  constructor(private prefix: string, private admins: string[] = []) {}

  async register(name: string, handler: Command): Promise<void> {
    if (name.length === 1) {
      throw new Error("Invalid command length, reserved for chain commands");
    }

    await this.unregister(name);
    await handler.open?.(this);
    this.#commands.set(name, handler);
  }

  async registerMod(source: string): Promise<void> {
    const mod: CommandModule = await import(source);
    const instance = await mod.create();
    if (instance.command === undefined) {
      throw new Error(
        `Invalid Command module ${source}, missing command prefix`,
      );
    }
    this.register(instance.command, instance);
  }

  async unregister(name: string): Promise<void> {
    await this.#commands.get(name)?.close?.();
    this.#commands.delete(name);
  }

  available(): string[] {
    let names = [];
    for (const name of this.#commands.keys()) {
      names.push(name);
    }
    names.sort();
    return names;
  }

  async close(): Promise<void> {
    for (const command of this.#commands.keys()) {
      await this.unregister(command);
    }
  }

  async handle(message: irc.Message): Promise<void> {
    if (!message.parameters[1].startsWith(this.prefix)) {
      return;
    }

    const text = message.parameters[1].substring(this.prefix.length);
    let [_, command_s, parameter] = text.match(/(.+?)(?:\s+(.*))?$/) ?? [
      undefined,
      undefined,
      undefined,
    ];

    if (
      command_s == undefined || text.trim().match(/^\W*$/)
    ) {
      // Ignore empty or punctuation only commands
      // Accidental triggers annoy people
      return;
    }

    const channel = message.direct ? "" : message.channel()!;
    const nickname = message.sender!.nickname;

    if (!this.#last_command.has(channel)) {
      this.#last_command.set(channel, new Map());
    }

    if (command_s.length === 1) {
      parameter = `${command_s} ${parameter}`;
      command_s = this.#last_command.get(channel)!.get(nickname);
    }

    if (command_s === "help") {
      // Special command, handled internally
      for (const command of this.#commands.keys()) {
        const desc = this.#commands.get(command)!.description;
        if (desc !== undefined) {
          await message.replyNotice(`${command}: ${desc}`, true);
        }
      }
      return;
    }

    const command = this.#commands.get(command_s ?? "");

    if (
      command_s === undefined ||
      command === undefined ||
      (command.admin && !this.isAdmin(nickname ?? ""))
    ) {
      message.replyNotice("Invalid command", true);
      console.info(
        `User ${nickname} in ${channel} attempted '${command_s}' with '${parameter}'`,
      );
      return;
    }

    command.handle?.(message, parameter);
    this.#last_command.get(channel)!.set(nickname, command_s);
  }

  async channelJoin(message: irc.Message, channel: string): Promise<void> {
    for (const command of this.#commands.keys()) {
      await this.#commands.get(command)?.channelJoin?.(message, channel);
    }
  }

  async channelLeave(message: irc.Message, channel: string): Promise<void> {
    for (const command of this.#commands.keys()) {
      await this.#commands.get(command)?.channelLeave?.(message, channel);
    }
  }

  async nickJoin(
    message: irc.Message,
    channel: string,
    nickname: string,
  ): Promise<void> {
    for (const command of this.#commands.keys()) {
      await this.#commands.get(command)?.nickJoin?.(message, channel, nickname);
    }
  }

  async nickLeave(
    message: irc.Message,
    channel: string,
    nickname: string,
  ): Promise<void> {
    for (const command of this.#commands.keys()) {
      await this.#commands
        .get(command)
        ?.nickLeave?.(message, channel, nickname);
    }
  }

  private isAdmin(nickname: string): boolean {
    return this.admins.includes(nickname);
  }
}
