/**
 * @file
 * @copyright 2020, Valentin Anger
 * @license ISC
 */

import * as irc from "../net/irc/mod.ts";
import { Commander } from "./command.ts";

const version = "0.1.0";

if (Deno.args.length == 0) {
  console.error("Insufficient number of arguments, expected config path");
}

let config: any;
try {
  const file = Deno.openSync(Deno.args[0]);
  config = JSON.parse(new TextDecoder().decode(Deno.readAllSync(file)));
} catch (e) {
  console.error("Failed to read configuration", e);
  Deno.exit(1);
}

let client = new irc.Client({
  nickname: config.nickname,
  hostname: config.hostname,
  port: config.port,
  use_tls: config.tls,
  join_channels: config.join_channels,
  nickserv_password: config.nickserv_password,
});

const commander = new Commander(config.command_prefix, config.admins);

for (const mod of config.load_mods) {
  const path = new URL(`./mods/${mod}.ts`, import.meta.url);
  await commander.registerMod(path.toString());
}

await commander.register("version", {
  description: "Get the bots version",
  async handle(message, _) {
    await message.replyNotice(version);
    return;
  },
});

try {
  await client.connect();

  for await (const message of client.messages()) {
    const channel = message.parameters[0];
    const nick = message.sender?.nickname;

    switch (message.command) {
      case "JOIN":
        if (nick == client.nickname) {
          await commander.channelJoin(message, channel);
        } else if (nick != undefined) {
          await commander.nickJoin(message, channel, nick);
        }
        break;
      case "KICK": {
        const target = message.parameters[1];
        if (target == client.nickname) {
          await commander.channelLeave(message, channel);
        } else if (target != undefined) {
          await commander.nickLeave(message, channel, target);
        }
        break;
      }
      case "PART":
        if (nick == client.nickname) {
          await commander.channelLeave(message, channel);
        } else if (nick != undefined) {
          await commander.nickLeave(message, channel, nick);
        }
        break;
      case "PRIVMSG":
        await commander.handle(message);
        break;
    }
  }
} catch (e) {
  if (e instanceof irc.IrcError) {
    console.error(e.toString());
  } else {
    throw e;
  }
} finally {
  client.disconnect();
  commander.close();
}
