/**
 * @file
 * @copyright 2020, Valentin Anger
 * @license ISC
 */

import * as irc from "../../net/irc/mod.ts";
import { Command } from "../command.ts";

type Color = "Blue" | "Green" | "Red" | "Yellow";
const colors: ReadonlyArray<Color> = [
  "Blue",
  "Green",
  "Red",
  "Yellow",
] as const;
type Value =
  | "0"
  | "1"
  | "2"
  | "3"
  | "4"
  | "5"
  | "6"
  | "7"
  | "8"
  | "9"
  | "S"
  | "R"
  | "D2"
  | "W"
  | "WD4";
const values: ReadonlyArray<Value> = [
  "0",
  "1",
  "2",
  "3",
  "4",
  "5",
  "6",
  "7",
  "8",
  "9",
  "S",
  "R",
  "D2",
  "W",
  "WD4",
] as const;
const wild_cards: ReadonlyArray<Value> = ["W", "WD4"] as const;

interface Card {
  color: Color;
  value: Value;
}

type Hand = Card[];

function cardString(
  card: Card,
  opts: { colored_wild?: boolean; no_background?: boolean } = {},
): string {
  return `${irc.bold()}${
    irc.color(
      !opts.colored_wild && wild_cards.includes(card.value)
        ? irc.Color.Default
        : card.color === "Blue"
        ? irc.Color.Blue
        : card.color === "Green"
        ? irc.Color.Green
        : card.color === "Red"
        ? irc.Color.Red
        : irc.Color.Yellow,
      opts.no_background ? undefined : irc.Color.Black,
    )
  }${card.value}${irc.reset()}`;
}

function parseCardString(card_string: string): Card | null {
  const suit_string = card_string.slice(0, 1).toUpperCase();
  const value_string = card_string.slice(1).toUpperCase();

  const suit = colors.find((c) => c[0] === suit_string);
  const value = values.find((v) => v === value_string);

  if (suit === undefined || value === undefined) return null;

  return {
    color: suit,
    value,
  };
}

function randomCard(): Card {
  return {
    color: colors[Math.floor(Math.random() * colors.length)],
    value: values[Math.floor(Math.random() * values.length)],
  };
}

function randomCards(n: number): Card[] {
  if (n < 0) throw new Error("Expected positive number");

  const cards = [];
  for (let i = 0; i < n; i++) cards.push(randomCard());

  return cards;
}

function handString(hand: Card[]): string {
  return hand
    .map((c) => cardString(c))
    .sort()
    .join(" ");
}

enum Mode {
  upgrade = "upgrade",
}

class Game {
  static initial_card_count = 30;
  static minimum_start_hand = 7;

  #current_card: Card = randomCard();
  #current_player: number = 0;
  #draw_count: number = 0;
  #hands: Map<string, Hand> = new Map();
  #modes: Mode[];
  #order: string[] = [];
  #running = false;

  game_ended = false;

  constructor(modes: Mode[] = []) {
    this.#modes = modes;
  }

  private advanceTurn() {
    this.#current_player++;
    this.#current_player %= this.#order.length;
  }
  private async ensureTurn(msg: irc.Message, player: string): Promise<boolean> {
    if (!this.#running) {
      await msg.replyNotice("The game is not running", true);
      return false;
    }
    if (this.game_ended) {
      await msg.replyNotice("The game has ended", true);
      return false;
    }
    if (player !== this.#order[this.#current_player]) {
      await msg.replyNotice("It is not your turn.", true);
      return false;
    }
    return true;
  }
  private async gameEndCondition(msg: irc.Message) {
    if (this.#order.length < 2) {
      await msg.replyNotice(
        `The game is over, remaining player: ${this.#order[0]}`,
      );
      this.game_ended = true;
      return true;
    }
    return false;
  }
  private async sendCurrentTurn(msg: irc.Message) {
    const player = this.#order[this.#current_player];
    await msg.replyNotice(
      `It's ${player}'s (${
        this.#hands.get(player)!.length
      }) turn. Current card: ${
        cardString(this.#current_card, { colored_wild: true })
      }` +
        (this.#draw_count > 0 ? ` Draw count: ${this.#draw_count}` : ""),
    );
  }
  private async finishedPlayers(msg: irc.Message) {
    const promises: Promise<void>[] = [];
    this.#order = this.#order.filter((player, i) => {
      if (this.#hands.get(player)!.length === 0) {
        this.#hands.delete(player);
        if (i < this.#current_player) this.#current_player--;
        promises.push(msg.replyNotice(`${player} has finished the game.`));
        return false;
      }
      return true;
    });
    await Promise.all(promises);

    if (this.#current_player >= this.#order.length) this.#current_player = 0;
  }

  async draw(msg: irc.Message, player: string) {
    if (!(await this.ensureTurn(msg, player))) return;

    this.#draw_count = Math.max(this.#draw_count, 1);

    const player_hand = this.#hands.get(player);
    this.#hands.set(player, player_hand!.concat(randomCards(this.#draw_count)));
    this.#draw_count = 0;

    this.advanceTurn();
    await this.sendPlayerHand(msg, player);
    await this.sendCurrentTurn(msg);
  }
  async play(msg: irc.Message, player: string, card_string: string) {
    if (!(await this.ensureTurn(msg, player))) return;

    const selected_card = parseCardString(card_string);

    if (selected_card === null) {
      await msg.replyNotice("Invalid card string. Examples: R2 YWD4 GD2", true);
      return;
    }

    const player_hand = this.#hands.get(player);
    const card_index = player_hand?.findIndex((c) => {
      // The color does not matter for wilds
      return (
        c.value == selected_card.value &&
        (wild_cards.includes(c.value) ? true : c.color == selected_card.color)
      );
    });
    if (card_index === undefined || card_index === -1) {
      await msg.replyNotice("You don't have this card.", true);
      return;
    }

    const valid = this.#draw_count > 0
      ? // Can only play draw cards
        this.#current_card.value === selected_card.value ||
        (this.#modes.includes(Mode.upgrade) &&
          this.#current_card.value === "D2" &&
          selected_card.value === "WD4")
      : this.#current_card.color === selected_card.color ||
        this.#current_card.value === selected_card.value ||
        wild_cards.includes(selected_card.value);

    if (!valid) {
      await msg.replyNotice("This card cannot be played.", true);
      return;
    }

    player_hand?.splice(card_index, 1);
    this.#current_card = selected_card;

    // Special effects
    switch (selected_card.value) {
      case "D2":
        this.#draw_count += 2;
        break;
      case "R":
        // Should act like Skip with two players left.
        // Not changing the #current_player forces the current player
        // to be the next player. [(A), B] -> [B, (A)]
        if (this.#order.length > 2) {
          this.#current_player = this.#order.length - this.#current_player - 1;
        }
        this.#order = this.#order.reverse();
        break;
      case "S":
        this.advanceTurn();
        break;
      case "WD4":
        this.#draw_count += 4;
        break;
    }

    this.advanceTurn();

    await this.finishedPlayers(msg);
    if (await this.gameEndCondition(msg)) {
      return;
    }
    await this.sendPlayerHand(msg, player);
    await this.sendCurrentTurn(msg);
  }
  async playerJoin(msg: irc.Message, player: string) {
    if (this.#running) {
      msg.replyNotice("The game is already running", true);
      return;
    }

    if (!this.#order.includes(player)) {
      this.#hands.set(player, []);
      this.#order.push(player);

      await msg.replyNotice(
        `${this.#order.length} player(s) waiting to play Uno${
          this.#modes.length > 0 ? " (" + this.#modes.join(",") + ")" : ""
        }.`,
      );
    }
  }
  async playerLeave(msg: irc.Message, player: string) {
    const index = this.#order.findIndex((u) => u === player);
    if (index >= 0) {
      this.#hands.delete(player);
      this.#order.splice(index, 1);

      if (this.#current_player > index) {
        this.#current_player -= 1; // Players shifted back, move pointer back
      } else if (this.#current_player >= this.#order.length) {
        this.#current_player = 0; // Wrap around if the player at the end of the list leaves
      }

      if (!this.#running) {
        await msg.replyNotice(
          `${this.#order.length} player(s) waiting to play Uno.`,
        );
      }
    }

    if (!this.#running || (await this.gameEndCondition(msg))) {
      return;
    }
    await this.sendCurrentTurn(msg);
  }
  async sendPlayerHand(msg: irc.Message, player: string) {
    if (!this.#hands.has(player)) return;
    await msg.replyNotice(
      `Your hand: ${handString(this.#hands.get(player)!)}`,
      true,
    );
  }
  async start(msg: irc.Message, player: string) {
    if (this.#running) {
      await msg.replyNotice("A game is already in progress.", true);
      return;
    }

    if (this.#order.length < 2) {
      await msg.replyNotice(
        "There aren't enough players to start the game",
        true,
      );
      return;
    }
    if (this.#order[0] !== player) {
      await msg.replyNotice(
        "Only the games initiator can start the game",
        true,
      );
      return;
    }

    shuffle(this.#order);

    // Deal hands
    const hand_count = Math.max(
      Math.floor(Game.initial_card_count / this.#order.length),
      Game.minimum_start_hand,
    );

    for (const player of this.#order) {
      this.#hands.set(player, randomCards(hand_count));
      await msg.client.sendNotice(
        player,
        `Your hand: ${handString(this.#hands.get(player)!)}`,
      );
    }

    this.#running = true;

    await this.sendCurrentTurn(msg);
  }
}

class Uno implements Command {
  readonly command: string = "uno";
  readonly description: string = "An Uno clone";

  #games: Map<string, Game> = new Map();

  async handle(
    message: irc.Message,
    argument: string | undefined,
  ): Promise<void> {
    if (message.direct) {
      await message.replyNotice("Uno does not work in direct chats");
      return;
    }
    const channel = message.channel()!;
    const player = message.sender!.nickname;

    const game_instance = this.#games.get(channel);

    const [action, parameter] = (argument ?? "")
      .split(" ")
      .filter((s) => s.length > 0);

    switch ((action ?? "").toLowerCase()) {
      case "j":
      case "join": {
        const game = game_instance ??
          new Game(parameter.split(",").filter((v) => v in Mode) as Mode[]);
        if (game_instance === undefined) {
          this.#games.set(channel, game);
        }
        await game.playerJoin(message, player);
        break;
      }
      case "l":
      case "leave":
        await game_instance?.playerLeave(message, player);
        break;
      case "p":
      case "play":
        if (game_instance === undefined) {
          await message.replyNotice("There is no active game", true);
          break;
        }
        if (parameter === undefined) {
          await message.replyNotice("Missing card to play", true);
          break;
        }
        await game_instance.play(message, player, parameter);
        break;
      case "d":
      case "draw":
        if (game_instance === undefined) {
          await message.replyNotice("There is no active game", true);
          break;
        }
        await game_instance.draw(message, player);
        break;
      case "c":
      case "cards":
        if (game_instance === undefined) {
          await message.replyNotice("There is no active game", true);
          break;
        }
        await game_instance.sendPlayerHand(message, player);
        break;
      case "s":
      case "start":
        if (game_instance === undefined) {
          await message.replyNotice("There is no active game", true);
          break;
        }
        await game_instance.start(message, player);
        break;
      case "h":
      case "help":
      default:
        for (
          const line of `
Uno: Invoke a command with \`uno <command name> …\` or short command \`<short command> …\`
join (j), leave (l), start (s): Join, leave and start the game
play (p) <card>, draw (d), cards (c): Play a card, draw a card and get your hand
Cards are of the format [color][value], like G3 for Green Three, YS for Yellow Skip and RW for Red Wild
`.trim().split("\n")
        ) {
          await message.replyNotice(
            line,
            true,
          );
        }
        break;
    }

    if (game_instance?.game_ended) {
      this.#games.delete(channel);
    }
  }

  async channelLeave(message: irc.Message, channel: string): Promise<void> {
    this.#games.delete(channel);
  }

  async nickLeave(
    message: irc.Message,
    channel: string,
    nickname: string,
  ): Promise<void> {
    this.#games.get(channel)?.playerLeave(message, nickname);
  }
}

export function create() {
  return new Uno();
}

function shuffle(array: any[]) {
  for (let i = Math.random() * array.length * 100; i > 0; i--) {
    const a = Math.floor(Math.random() * array.length);
    const b = Math.floor(Math.random() * array.length);
    [array[a], array[b]] = [array[b], array[a]];
  }
}
